// Copyright 1994 by Jon Dart.  All Rights Reserved.

#ifndef _MATERIAL_H
#define _MATERIAL_H

#include "chess.h"
#include "debug.h"

class Material
{
	friend class Board;
public:
	Material():
        info(0), total(0), level(0)
	{
	}
	
        enum {KP = 0x100001};
        enum {KBP = 0x100101};
        enum {KBN = 0x100110};
        enum {KRB = 0x101100};
        enum {KRN = 0x101010};
        enum {KRBN = 0x101110};
        enum {KNN = 0x100020};
        enum {KRR = 0x102000};
        enum {KB  = 0x100100};
        enum {KN  = 0x100010};
        enum {KR  = 0x101000};
        enum {KQ  = 0x110000};

	enum {PieceMask = 0x1ffff0};
        
        FORCEINLINE void addPiece(const PieceType p)
        {
           info += masks[(int)p];
           total += PieceValue(p);
           level += levels[(int)p];
           count++;
           ASSERT(level >=0);
        }

	FORCEINLINE void removePiece(const PieceType p)
        {
          info -= masks[(int)p];
          total -= PieceValue(p);
          level -= levels[(int)p];
          count--;
          ASSERT(level >=0);
        }

        FORCEINLINE void addPawn()
        {
           info += masks[(int)Pawn];
           total += PAWN_VALUE;
           ++count; 
        }

	FORCEINLINE void removePawn()
        {
          info -= masks[(int)Pawn];
          total -= PAWN_VALUE;
          --count;
        }

	// return the total material value:
	uint32 value() const
	{
		return total;
	}
	
	uint32 infobits() const
	{
		return info;
	}
	
	// return the total number of pieces + pawns
	int men() const {
	  return count;
        }
        
	// return the number of pieces (excluding the King)
	int pieceCount() const {
	  return count-pawnCount()-1;
        }
        
        // return the number of pawns
        int pawnCount() const
        {
                return (info & 0xf);
        }
        
        int queenCount() const
        {
                return (info & 0xf0000) >> 16;
        }

        int hasQueen() const
        {
                return (info & 0xf0000);
        }

        int rookCount() const
        {
                return (info & 0xf000) >> 12;
        }
        
        int hasRook() const
        {
                return (info & 0xf000);
        }

        int knightCount() const
        {
                return (info & 0xf0) >> 4;
        }
        
        int hasKnight() const
        {
                 return (info & 0xf0);
        }

        int bishopCount() const
        {
                return (info & 0xf00) >> 8;
        }

        int hasBishop() const
        {
                 return (info & 0xf00);
        }

	// True if only king and pawns:
	bool noPieces() const
	{
		return (info & (0xffffffff-0x100000-0xf)) == 0;
	}
        
        // True if no pawns
        bool noPawns() const
        {
                return (info & 0xf) == 0;
        }
	
	// True if bare king:
	bool kingOnly() const
	{
		return (info & 0xfffff) == 0;
	}
        
        // This provides a rough measure of the material
        // (used mostly for king safety calculation)
        int materialLevel() const {
	  return (level > 31) ? 31 : level;
        }
	
private:
        void clear() {
          info = total = level = count = 0;
        }

	int32 info;
	uint32 total;
        int32 level;
        int count;
        static const int32 masks[8];
        static const int32 masks2[8];
        static const int32 levels[8];
};

#endif

